""" Typing shortcuts for Hekr API """
from typing import List, Dict, Union, Any, AnyStr

DecodedDataType = Dict[str, Any]
DatagramType = AnyStr
CommandSearchIdentifier = Union[str, int]
AnyNumber = Union[int, float, None]